/***********************************************************
*   MPD Heave Rig Simulator
*   Linear version
*   Mikael nestad 2013
*/

#define S_FUNCTION_NAME  mpdriglinsfcn
#define S_FUNCTION_LEVEL 2

#include "simstruc.h"
#include "math.h"

/***********************************************************
*   Constant definitions
*   The constant are passed from Simulink
*   (Ctrl-U) when the mask is selected.
*/

#define C_N                 100 // Number of discretization elements
#define NPARAMS             28
#define C_A                 (mxGetPr(ssGetSFcnParam(S,0)))[0]
#define C_RHO_REF           (mxGetPr(ssGetSFcnParam(S,1)))[0]
#define C_P_REF             (mxGetPr(ssGetSFcnParam(S,2)))[0]
#define C_F1                (mxGetPr(ssGetSFcnParam(S,3)))[0]
#define C_F2                (mxGetPr(ssGetSFcnParam(S,4)))[0]
#define C_C                 (mxGetPr(ssGetSFcnParam(S,5)))[0]
#define C_PBAR              (mxGetPr(ssGetSFcnParam(S,6)))[0] 
#define C_UBAR              (mxGetPr(ssGetSFcnParam(S,7)))[0] 
#define C_P_OUT             (mxGetPr(ssGetSFcnParam(S,8)))[0]
#define C_PL                (mxGetPr(ssGetSFcnParam(S,9)))[0] 
#define C_T                 (mxGetPr(ssGetSFcnParam(S,10)))[0]
#define C_CH_RHO_L          (mxGetPr(ssGetSFcnParam(S,11)))[0]  
#define C_GAMMA             (mxGetPr(ssGetSFcnParam(S,12)))[0] 
#define C_CH                (mxGetPr(ssGetSFcnParam(S,13)))[0]  
#define C_CAD               (mxGetPr(ssGetSFcnParam(S,14)))[0]  
#define C_CAD2              (mxGetPr(ssGetSFcnParam(S,15)))[0]  
#define C_CF1               (mxGetPr(ssGetSFcnParam(S,16)))[0]  
#define C_CF2               (mxGetPr(ssGetSFcnParam(S,17)))[0]  
#define C_L                 (mxGetPr(ssGetSFcnParam(S,18)))[0]  
#define C_STATIC_OUTPUTS    (mxGetPr(ssGetSFcnParam(S,19)))[0]  
#define C_CHB               (mxGetPr(ssGetSFcnParam(S,22)))[0]  
#define C_CAB               (mxGetPr(ssGetSFcnParam(S,23)))[0]  
#define C_CAW               (mxGetPr(ssGetSFcnParam(S,24)))[0]  
#define C_K1                (mxGetPr(ssGetSFcnParam(S,25)))[0]  // reserved for future parameters
#define C_K2                (mxGetPr(ssGetSFcnParam(S,26)))[0]  // reserved for future parameters  
#define C_BETA              (mxGetPr(ssGetSFcnParam(S,27)))[0] 
#define C_K                 (C_C*C_C*C_RHO_REF-C_P_REF)
#define C_G                 9.81 

#define MIN(a,b) (((a)<(b))?(a):(b))
#define MAX(a,b) (((a)>(b))?(a):(b))

static real_T solvexstar(SimStruct *S, real_T alphaN, real_T Gu, real_T q_b);
static void alpha_interp(real_T* yvalues, real_T* xvalues, real_T* yinterp, int M, int P, real_T dx, real_T alphaStart);
static void beta_interp(real_T* yvalues, real_T* xvalues, real_T* yinterp, int M, int P, real_T dx, real_T betaStart);
static void interp(real_T* yvalues, real_T* xvalues, real_T* yinterp, int M, int P, real_T dx);

/***********************************************************
*   Initializing all the vectors involved
*/

#define MDL_INITIAL_SIZES
static void mdlInitializeSizes(SimStruct *S)
{    
    int i;
    // number of input ports
    int_T nInputPorts  = 3;        
    
    // number of output ports
    int_T nOutputPorts = (int)C_STATIC_OUTPUTS + (int)mxGetN(ssGetSFcnParam(S,20)); 
    
    // number of expected parameters
    ssSetNumSFcnParams(S, NPARAMS);    
    
    // number of continuous states
    ssSetNumContStates(S, C_N*2+1);     
    
    // number of discrete states
    ssSetNumDiscStates(S, 0);

    if (!ssSetNumInputPorts(S, nInputPorts)) return;   
    ssSetInputPortWidth(S, 0, 1);
    ssSetInputPortWidth(S, 1, 1);
    ssSetInputPortWidth(S, 2, 1);

    ssSetInputPortDirectFeedThrough(S, 1, 1);   // we need the piston velocity in the output
    ssSetInputPortDirectFeedThrough(S, 2, 1);   // we need the pump flow in the output
    ssSetNumIWork(S, 1);

    if (!ssSetNumOutputPorts(S, nOutputPorts)) return; 
    for(i=0; i<nOutputPorts; i++){
        ssSetOutputPortWidth(S, i, 1);			// setting the width of all the pressure outputs
    }    
}

#define MDL_INITIALIZE_SAMPLE_TIMES
static void mdlInitializeSampleTimes(SimStruct *S)
{
    ssSetSampleTime(S, 0, CONTINUOUS_SAMPLE_TIME);
    ssSetOffsetTime(S, 0, 0.0);
    ssSetModelReferenceSampleTimeDefaultInheritance(S);
    
}

/***********************************************************
*   Initializing the states. Except the choke opening, which is initialized
*   in the output part (since some of the inputs are needed)
*/

#define MDL_INITIALIZE_CONDITIONS
static void mdlInitializeConditions(SimStruct *S)
{
    real_T *x0 = ssGetContStates(S);
    real_T q_b = *(ssGetInputPortRealSignalPtrs(S,2))[0];

    int j;
    double P[C_N+1], U[C_N+1], deltax;

	ssSetIWorkValue(S, 0, 1);

    deltax = C_L/C_N;
    
    for(j=0; j<C_N+1; j++){
        P[j] = (C_K+C_PL)*exp(C_G*sin(C_GAMMA)/pow(C_C,2)*(C_L-(real_T)j*deltax))-C_K; 
        U[j] = 0;
    }   
        
    for(j=0; j<C_N; j++){
        x0[j] = (P[j+1]-C_PBAR)/(C_RHO_REF*C_C)+(U[j+1]-C_UBAR);       // alphas
        x0[j+C_N] = -(P[j]-C_PBAR)/(C_RHO_REF*C_C)+(U[j]-C_UBAR);       // betas
    }

}

/***********************************************************
*   Handling the outputs. Some values have to be recalculated, 
*   since they are not states of the system, and hence not 
*   available in the output section of the S-function
*/

#define MDL_OUTPUTS
static void mdlOutputs(SimStruct *S, int_T tid)
{
    real_T	*y    = ssGetOutputPortRealSignal(S,0);
    real_T	*x    = ssGetContStates(S);
    real_T *G = mxGetPr(ssGetSFcnParam(S,21));
    real_T U0, alphaStart, q_in, u, Gu, xstar, betaStart, q_out, q_c, p_c, deltaU;
    int i,j, numOutputs;
    double *output_alphas, *output_betas;
 
    real_T v_d = *(ssGetInputPortRealSignalPtrs(S,1))[0];
    real_T q_b = *(ssGetInputPortRealSignalPtrs(S,2))[0];
    int numG = (int)mxGetN(ssGetSFcnParam(S,21));
    
    // allocating memory for pipe pressures
    numOutputs = (int)mxGetN(ssGetSFcnParam(S,20));
    output_alphas = malloc(numOutputs * sizeof(double));
    output_betas = malloc(numOutputs * sizeof(double));
    
    // allocating memory for pipe pressures
    // (since the number of these can be changed from Simulink)
    if (ssGetIWorkValue(S, 0) == 1) {
        // setting initial choke opening
        Gu = q_b/sqrt((C_PL-C_P_OUT)/C_RHO_REF);

        deltaU = 1/((double)numG-1);

        if(Gu >= G[numG-1]){
            x[2*C_N] = 1; // u
        } else if(Gu <= 0){
             x[2*C_N] = 0;
        }else{
             // interpolation
             j=0;
             while(G[j] <= Gu){
                j++;
                if(j == numG-1) break;
             }
             x[2*C_N] = deltaU*(Gu-G[j-1])/(G[j]-G[j-1])+deltaU*(j-1);
        }
        ssSetIWorkValue(S, 0, 0);
       printf("Initial pump: %f\n", q_b);

       printf("Initial choke opening: %f\n", x[2*C_N]);
    }
	

    // The pipe inlet flow (from the piston)
    U0 = -v_d*(C_CAD-C_CAD2)/C_A;                   // ignoring compressibility
    //U0 = -x[2*C_N+1];                             // uncomment this for compressibility effects (needs tuning)
	
	// Calculating the lower boundary
    alphaStart = 2*(U0-C_UBAR)-x[C_N];            
    q_in = C_A*(C_UBAR+(alphaStart+x[C_N])/2);    
    y[5] = C_A*U0;
    
    // Actual choke opening.
    u = x[2*C_N];
    y[3] = u;
    
    // The choke coefficient is found by interpolating the choke characteristic
    deltaU = 1/((double)mxGetN(ssGetSFcnParam(S,21))-1);
    interp((mxGetPr(ssGetSFcnParam(S,21))), &u, &Gu, (int)mxGetN(ssGetSFcnParam(S,21)), 1, deltaU);
    
    // Compute xstar (q_c/A)
    xstar = solvexstar(S, x[C_N-1], Gu, q_b);
    betaStart = 2*(xstar-q_b/C_A-C_UBAR)-x[C_N-1];
    q_out = C_A*(C_UBAR+(x[C_N-1]+betaStart)/2);
    q_c = q_out+q_b;
    y[1] = q_c;
    y[6] = q_out;
    
    // Pressure at inlet of pipe
    y[0] = C_RHO_REF*C_C*(alphaStart-x[C_N])/2 + C_PBAR;

    // Compute pressure at choke inlet
    if (Gu>0){
        p_c = C_P_OUT+(xstar/fabs(xstar))*C_CH_RHO_L*pow((q_c/Gu),2);
    }else{
        p_c = C_RHO_REF*C_C*(x[C_N-1]-betaStart)/2 + C_PBAR;
    }
    y[2] = p_c;
    
    // Bottom hole pressure
    y[4] = y[0] + C_CF1*v_d +C_CF2*v_d*fabs(v_d) + C_RHO_REF*C_G*C_CH;
   
    // Intermediate pressures (interpolated)
    alpha_interp(x, (mxGetPr(ssGetSFcnParam(S,20))), output_alphas, C_N, numOutputs,  C_L/C_N, alphaStart);
    beta_interp(&x[C_N+1], (mxGetPr(ssGetSFcnParam(S,20))), output_betas, C_N, numOutputs, C_L/C_N, betaStart); 
    
    for(i=0; i<numOutputs; i++){
        y[(int)C_STATIC_OUTPUTS+i] = C_RHO_REF*C_C*(output_alphas[i]-output_betas[i])/2+C_PBAR;
    }
    
    free(output_alphas);
    free(output_betas);
}


/***********************************************************
*   Handling the derivatives.
*/

#define MDL_DERIVATIVES
static void mdlDerivatives(SimStruct *S)
{   
    real_T *dx     = ssGetdX(S);
    real_T *x = ssGetContStates(S);
    real_T uc = *(ssGetInputPortRealSignalPtrs(S,0))[0];
    real_T v_d = *(ssGetInputPortRealSignalPtrs(S,1))[0];
    real_T q_b = *(ssGetInputPortRealSignalPtrs(S,2))[0];
  
    real_T U0, alphaStart, betaStart, deltax, dalphadx[C_N], Gu, u, xstar, dbetadx[C_N], uf[C_N+1], fr[C_N+1], du, deltaU;
    int i;

    u = x[2*C_N];
    
   // The pipe inlet flow (from the piston)
    U0 = -v_d*(C_CAD-C_CAD2)/C_A;                    // ignoring compressibility
    // U0 = x[2*C_N+1];                              // uncomment this for compressibility effects (needs tuning)
        
    // Lower boundary
    alphaStart = 2*(U0-C_UBAR)-x[C_N]; 
    
    // The choke coefficient is found by interpolating the choke characteristic
    deltaU = 1/((double)mxGetN(ssGetSFcnParam(S,21))-1);
    interp((mxGetPr(ssGetSFcnParam(S,21))), &u, &Gu, (int)mxGetN(ssGetSFcnParam(S,21)), 1, deltaU);
    
    // Compute xstar (q_c/A)
    xstar = solvexstar(S, x[C_N-1], Gu, q_b);
    betaStart= 2*(xstar-q_b/C_A-C_UBAR)-x[C_N-1];    // (44)

    // Compute the process flux
    deltax = C_L/C_N;   
    
    dalphadx[0] = (x[0] - alphaStart)/deltax;
    dbetadx[C_N-1] = (betaStart - x[2*C_N-1])/deltax;
    
    for(i=1; i<C_N; i++){
        dalphadx[i] = (x[i]-x[i-1])/deltax;
        dbetadx[i-1] = (x[C_N+i]-x[C_N+i-1])/deltax;
    }  
    
    uf[0] = U0;
    uf[C_N] = C_UBAR + (x[C_N-1]+betaStart)/2;
    fr[0] = C_F1*uf[0] + C_F2*fabs(uf[0])*uf[0];
    fr[C_N] = C_F1*uf[C_N] + C_F2*fabs(uf[C_N])*uf[C_N];
    
    for(i=1; i<C_N; i++){    
        uf[i] = C_UBAR + (x[i-1]+x[C_N+i])/2;
        fr[i] = C_F1*uf[i] + C_F2*fabs(uf[i])*uf[i];
    }
    
    for(i=0; i<C_N; i++){ 
        // Process time derivative        
        dx[i] = -C_C*dalphadx[i] - fr[i+1] - C_G*sin(C_GAMMA); // dalphadt
        dx[i+C_N] = C_C*dbetadx[i] - fr[i] - C_G*sin(C_GAMMA); // dbetadt
    }

    // We assume the dynamics of the choke opening to follow a first order
    // filter, which is tuned a little "faster" than the given opening
    // time. The rate is then saturated according to the given opening
    // time.

    du = (-x[2*C_N]+uc)/(0.1*C_T);
    du = MAX(MIN(du,1/C_T),-1/C_T);
    
    // The opening needs to be saturated between 0 and 1.
    if (u>=1 && du>0){
        du=0;
    }
    
    if (u<=0 && du<0){
        du=0;
    }

    dx[2*C_N] = du;
} 


static void mdlTerminate(SimStruct *S)
{
    UNUSED_ARG(S); /* unused input argument */
}


/***********************************************************
*   Solving the quadratic equation related to the choke boundary.
*   - alphaN is the alpha value for x=l
*   - Gu is the choke coefficient
*   - q_b is the pump flow [m3/s]
*/
static real_T solvexstar(SimStruct *S, real_T alphaN, real_T Gu, real_T q_b){
    real_T xstar = 0;
    real_T K1, K2, K3, sol1, sol2;
        
    if(Gu > 0){
        
        K1 = -pow(C_A,2)*C_RHO_REF/pow(Gu,2);
        K2 = -C_RHO_REF*C_C;
        K3 = C_PBAR-C_P_OUT+C_RHO_REF*C_C*(alphaN+C_UBAR+q_b/C_A);

        sol1 = (-K2+sqrt(pow(K2,2)-4*K1*K3))/(2*K1);
        sol2 = (-K2-sqrt(pow(K2,2)-4*K1*K3))/(2*K1);
        
        xstar = -1;
        if(sol1 >= 0)
            xstar = sol1;
        else if(sol2 >= 0)
            xstar = sol2;
        
        if(xstar < 0){
            K1 = -1*K1;    
            sol1 = (-K2+sqrt(pow(K2,2)-4*K1*K3))/(2*K1);
            sol2 = (-K2-sqrt(pow(K2,2)-4*K1*K3))/(2*K1);
            
            if(sol1 <= 0)
                xstar = sol1;
            else if(sol2 <= 0)
                xstar = sol2;
        }
    }

    return xstar;
}

/***********************************************************
*   Linear interpolation of evenly spaced values
*   - yvalues are the known y-values (M elements, with a spacing dx from x=0)
*   - xvalues are the desired x-values (P elements)
*/
static void interp(real_T* yvalues, real_T* xvalues, real_T* yinterp, int M, int P, real_T dx){
	int i, lowindex;
	real_T ylow, yhi;
	for(i=0; i<P; i++){
		lowindex = (int)floor(xvalues[i]/dx);        
		ylow = yvalues[lowindex];
		yhi = yvalues[lowindex+1];
		yinterp[i] = (xvalues[i]-lowindex*dx)*(yhi-ylow)/dx+ylow;
	}	
}


/***********************************************************
*   Linear interpolation of evenly spaced values when one value is appended
*   to the start of the source vector (see interp())
*/
static void alpha_interp(real_T* yvalues, real_T* xvalues, real_T* yinterp, int M, int P, real_T dx, real_T alphaStart){
    int i, lowindex;
	real_T ylow, yhi;
    for(i=0; i<P; i++){
        if(xvalues[i] <= 0 && xvalues[i] < dx){
            yinterp[i] = xvalues[i]*(yvalues[0]-alphaStart)/dx+alphaStart;
        } else {
            lowindex = (int)floor((xvalues[i]-dx)/dx);
            ylow = yvalues[lowindex];
            yhi = yvalues[lowindex+1];
            yinterp[i] = (xvalues[i]-(lowindex+1)*dx)*(yhi-ylow)/dx+ylow;
        }
    }
}

/***********************************************************
*   Linear interpolation of evenly spaced values when one value is appended
*   to the end of the source vector (see interp())
*/
static void beta_interp(real_T* yvalues, real_T* xvalues, real_T* yinterp, int M, int P, real_T dx, real_T betaStart){
    int i, lowindex;
	real_T xlow, xhi, ylow, yhi;
    for(i=0; i<P; i++){
        if(xvalues[i] >= (M-1)*dx && xvalues[i] <= M*dx){
            yinterp[i] = (xvalues[i]-(M-1)*dx)*(betaStart-yvalues[M-1])/dx+yvalues[M-1];
        } else {
            lowindex = (int)floor(xvalues[i]/dx);        
            ylow = yvalues[lowindex];
            yhi = yvalues[lowindex+1];
            yinterp[i] = (xvalues[i]-lowindex*dx)*(yhi-ylow)/dx+ylow;
        }
    }
}

#ifdef MATLAB_MEX_FILE /* Is this file being compiled as a MEX-file? */
#include "simulink.c" /* MEX-file interface mechanism */
#else
#include "cg_sfun.h" /* Code generation registration function */
#endif
