/***********************************************************
*   MPD Heave Rig Observer
*   Mikael nestad 2013 
*   Based on a paper by Ole Morten Aamo
*/


#define S_FUNCTION_NAME  mpdrigobssfcn
#define S_FUNCTION_LEVEL 2

#include "simstruc.h"
#include "math.h"


/***********************************************************
*   Constant definitions
*   The constant are passed from Simulink
*   (Ctrl-U) when the mask is selected.
*/

#define C_N                 20
#define NPARAMS             24
#define C_A                 (mxGetPr(ssGetSFcnParam(S,0)))[0]
#define C_RHO_REF           (mxGetPr(ssGetSFcnParam(S,1)))[0]
#define C_P_REF             (mxGetPr(ssGetSFcnParam(S,2)))[0]
#define C_F1                (mxGetPr(ssGetSFcnParam(S,3)))[0]
#define C_F2                (mxGetPr(ssGetSFcnParam(S,4)))[0]
#define C_C                 (mxGetPr(ssGetSFcnParam(S,5)))[0]
#define C_PBAR              (mxGetPr(ssGetSFcnParam(S,6)))[0] 
#define C_UBAR              (mxGetPr(ssGetSFcnParam(S,7)))[0] 
#define C_PL                (mxGetPr(ssGetSFcnParam(S,8)))[0] 
#define C_L                 (mxGetPr(ssGetSFcnParam(S,9)))[0]  
#define C_CH_RHO_L          (mxGetPr(ssGetSFcnParam(S,10)))[0]  
#define C_GAMMA             (mxGetPr(ssGetSFcnParam(S,11)))[0] 
#define C_CH                (mxGetPr(ssGetSFcnParam(S,12)))[0]  
#define C_CAD               (mxGetPr(ssGetSFcnParam(S,13)))[0]  
#define C_CAD2              (mxGetPr(ssGetSFcnParam(S,14)))[0]  
#define C_CF1               (mxGetPr(ssGetSFcnParam(S,15)))[0]  
#define C_CF2               (mxGetPr(ssGetSFcnParam(S,16)))[0]  
#define C_STATIC_OUTPUTS    (mxGetPr(ssGetSFcnParam(S,17)))[0]  
#define C_K                 (C_C*C_C*C_RHO_REF-C_P_REF)
#define C_G                 9.81 
#define C_D1               (mxGetPr(ssGetSFcnParam(S,21)))[0]  
#define C_D2               (mxGetPr(ssGetSFcnParam(S,22)))[0]  
#define C_OMEGA             (mxGetPr(ssGetSFcnParam(S,23)))[0]  

#define MIN(a,b) (((a)<(b))?(a):(b))
#define MAX(a,b) (((a)>(b))?(a):(b))

static void alpha_interp(real_T* yvalues, real_T* xvalues, real_T* yinterp, int M, int P, real_T dx, real_T alphaStart);
static void beta_interp(real_T* yvalues, real_T* xvalues, real_T* yinterp, int M, int P, real_T dx, real_T betaStart);
static void interp(real_T* yvalues, real_T* xvalues, real_T* yinterp, int M, int P, real_T dx);

#define MDL_INITIAL_SIZES
static void mdlInitializeSizes(SimStruct *S)
{    
    int i;
    // number of input ports
    int_T nInputPorts  = 2;             
    
    // number of output ports
    int_T nOutputPorts = (int)C_STATIC_OUTPUTS + (int)mxGetN(ssGetSFcnParam(S,18));
    
    // number of expected parameters
    ssSetNumSFcnParams(S, NPARAMS);    
    
    // number of continuous states
    ssSetNumContStates(S, C_N*2+3);
    
    // number of discrete states
    ssSetNumDiscStates(S, 0);           

    if (!ssSetNumInputPorts(S, nInputPorts)) return;   
    ssSetInputPortWidth(S, 0, 1);
    ssSetInputPortWidth(S, 1, 1);
    
    ssSetInputPortDirectFeedThrough(S, 0, 1);   // we need the outlet flow in the output
    ssSetInputPortDirectFeedThrough(S, 1, 1);   // we need the outlet pressure in the output

    if (!ssSetNumOutputPorts(S, nOutputPorts)) return; 
    for(i=0; i<nOutputPorts; i++){
        ssSetOutputPortWidth(S, i, 1);
    }    
}

#define MDL_INITIALIZE_SAMPLE_TIMES
static void mdlInitializeSampleTimes(SimStruct *S)
{
    ssSetSampleTime(S, 0, CONTINUOUS_SAMPLE_TIME);
    ssSetOffsetTime(S, 0, 0.0);
    ssSetModelReferenceSampleTimeDefaultInheritance(S);
    
}


/***********************************************************
*   Initializing the states. Except the choke opening, which is initialized
*   in the output part (since some of the inputs are needed)
*/
#define MDL_INITIALIZE_CONDITIONS
static void mdlInitializeConditions(SimStruct *S)
{
    real_T *x0 = ssGetContStates(S);
    int j;
    double P[C_N+1], U[C_N+1], deltax;

    deltax = C_L/C_N;
    
    for(j=0; j<C_N+1; j++){
        P[j] = (C_K+C_PL)*exp(C_G*sin(C_GAMMA)/pow(C_C,2)*(C_L-(real_T)j*deltax))-C_K;
        U[j] = 0;
    }   
        
    for(j=0; j<C_N; j++){
        x0[j] = (P[j+1]-C_PBAR)/(C_RHO_REF*C_C)+(U[j+1]-C_UBAR);        // alphas
        x0[j+C_N] = -(P[j]-C_PBAR)/(C_RHO_REF*C_C)+(U[j]-C_UBAR);       // betas
    }
    
    x0[2*C_N+1] = 0.01;
    x0[2*C_N+2] = 0.01;
}


/***********************************************************
*   Handling the outputs. Some values have to be recalculated, 
*   since they are not states of the system, and hence not 
*   available in the output section of the S-function
*/
#define MDL_OUTPUTS
static void mdlOutputs(SimStruct *S, int_T tid)
{
    real_T	*y    = ssGetOutputPortRealSignal(S,0);
    real_T	*x    = ssGetContStates(S);
    real_T U0, alphaStart, betaStart;
    int i, j, numOutputs;
    double *output_alphas, *output_betas;
	
    real_T p_plant = *(ssGetInputPortRealSignalPtrs(S,1))[0];
    real_T q_plant = *(ssGetInputPortRealSignalPtrs(S,0))[0];
	
    real_T v_d = x[2*C_N+2];
    
    // allocating memory dynamically for pipe pressures 
    // (since the number of these can be changed from Simulink)
    numOutputs = (int)mxGetN(ssGetSFcnParam(S,18));
    output_alphas = malloc(numOutputs * sizeof(double));
    output_betas = malloc(numOutputs * sizeof(double));
	
    // Compute q_in
    U0 = v_d*(C_CAD2-C_CAD)/C_A;                  // ignoring compressibility
    alphaStart = 2*(U0-C_UBAR)-x[C_N];

    y[4] = C_A*U0;

    betaStart= -1/(C_RHO_REF*C_C)*(p_plant-C_PBAR) + q_plant/C_A;  
    y[3] = C_A*(C_UBAR+(x[C_N-1]+betaStart)/2);
	
    // Pressure at outlet of pipe
    y[2] = C_RHO_REF*C_C*(x[C_N-1]-betaStart)/2 + C_PBAR;
    
    // Pressure at inlet of pipe
    y[0] = C_RHO_REF*C_C*(alphaStart-x[C_N])/2 + C_PBAR;
    
    // Bottom hole pressure
    y[1] = y[0] + C_CF1*v_d +C_CF2*v_d*fabs(v_d) + C_RHO_REF*C_G*C_CH;
   
    // Intermediate pressures (interpolated)
    alpha_interp(x, (mxGetPr(ssGetSFcnParam(S,18))), output_alphas, C_N, numOutputs,  C_L/C_N, alphaStart);
    beta_interp(&x[C_N+1], (mxGetPr(ssGetSFcnParam(S,18))), output_betas, C_N, numOutputs, C_L/C_N, betaStart); 
    
    y[5] = x[2*C_N+2];
            
    for(i=0; i<numOutputs; i++){
        y[(int)C_STATIC_OUTPUTS+i] = C_RHO_REF*C_C*(output_alphas[i]-output_betas[i])/2+C_PBAR;
    }
    
    free(output_alphas);
    free(output_betas);
}


/***********************************************************
*   Handling the derivatives
*/
#define MDL_DERIVATIVES
static void mdlDerivatives(SimStruct *S)
{   
    real_T *dx     = ssGetdX(S);
    real_T *x = ssGetContStates(S);

    real_T v_d = x[2*C_N+2];
    
    real_T k1, q_end_est, p_end_est, q_plant, p_plant, u_err, v_err, u_est, v_est, u_plant, v_plant, alphaCorr, betaCorr, p1corr[C_N+1], p2corr[C_N+1], xcoords[C_N+1];
    real_T U0, alphaStart, betaStart, deltax, dalphadx[C_N], dbetadx[C_N], uf[C_N+1], fr[C_N+1];
    int i,j;    
    
    // Inlet boundary
    U0 = -v_d*(C_CAD-C_CAD2)/C_A;
    alphaStart = 2*(U0-C_UBAR)-x[C_N];  

    p_plant = *(ssGetInputPortRealSignalPtrs(S,1))[0];
    q_plant = *(ssGetInputPortRealSignalPtrs(S,0))[0];

    betaStart= -1/(C_RHO_REF*C_C)*(p_plant-C_PBAR) + q_plant/C_A;
    
    deltax = C_L/C_N;   
    
    for(j=0; j<C_N+1; j++){
        xcoords[j] = j*C_L/C_N;
    }
    
    interp((mxGetPr(ssGetSFcnParam(S,19))), xcoords, p1corr, (int)mxGetN(ssGetSFcnParam(S,19)), C_N+1, C_L/(int)(mxGetN(ssGetSFcnParam(S,19))-1));
    interp((mxGetPr(ssGetSFcnParam(S,20))), xcoords, p2corr, (int)mxGetN(ssGetSFcnParam(S,20)), C_N+1, C_L/(int)(mxGetN(ssGetSFcnParam(S,20))-1));

    q_end_est = C_A*(C_UBAR+(x[C_N-1]+betaStart)/2);
    p_end_est = C_RHO_REF*C_C/2*(x[C_N-1]-betaStart)+C_PBAR;
    
    k1 = C_L*C_F1/(2*C_RHO_REF*C_C);
    
    u_est = 0.5*(q_end_est+(C_A/(C_RHO_REF*C_C))*(p_end_est+C_RHO_REF*C_G*sin(C_GAMMA)*C_L))*exp(k1);
    v_est = 0.5*(q_end_est-(C_A/(C_RHO_REF*C_C))*(p_end_est+C_RHO_REF*C_G*sin(C_GAMMA)*C_L))*exp(-k1);
 
    u_plant = 0.5*(q_plant+(C_A/(C_RHO_REF*C_C))*(p_plant+C_RHO_REF*C_G*sin(C_GAMMA)*C_L))*exp(k1);
    v_plant = 0.5*(q_plant-(C_A/(C_RHO_REF*C_C))*(p_plant+C_RHO_REF*C_G*sin(C_GAMMA)*C_L))*exp(-k1);
 
    u_err = u_plant-u_est;
    v_err = v_plant-v_est;
    
    dalphadx[0] = (x[0] - alphaStart)/deltax;
    dbetadx[C_N-1] = (betaStart - x[2*C_N-1])/deltax;
    
    for(i=1; i<C_N; i++){
        dalphadx[i] = (x[i]-x[i-1])/deltax;
        dbetadx[i-1] = (x[C_N+i]-x[C_N+i-1])/deltax;
    }  
    
    uf[0] = U0;
    uf[C_N] = C_UBAR + (x[C_N-1]+betaStart)/2;
    fr[0] = C_F1*uf[0] + C_F2*fabs(uf[0])*uf[0];
    fr[C_N] = C_F1*uf[C_N] + C_F2*fabs(uf[C_N])*uf[C_N];
    
    for(i=1; i<C_N; i++){    
        uf[i] = C_UBAR + (x[i-1]+x[C_N+i])/2;
        fr[i] = C_F1*uf[i] + C_F2*fabs(uf[i])*uf[i];
    }    
    
    // Compute the process flux    
    for(i=0; i<C_N; i++){ 
        // observer injections
        alphaCorr = 2*p1corr[i+1]*u_err*exp(-k1*(i+1)/C_N)/C_A;
        betaCorr = 2*p2corr[i]*u_err*exp(k1*i/C_N)/C_A;

        // Process time derivative        
        dx[i] = -C_C*dalphadx[i] - fr[i+1] - C_G*sin(C_GAMMA) + alphaCorr;  // dalphadt
        dx[i+C_N] = C_C*dbetadx[i] - fr[i] - C_G*sin(C_GAMMA) + betaCorr;   // dbetadt
    }

    dx[2*C_N+1] = -C_OMEGA*x[2*C_N+2]  + C_D1 *u_err;
    dx[2*C_N+2] =  C_OMEGA*x[2*C_N+1]  + C_D2 *u_err;
} 

static void mdlTerminate(SimStruct *S)
{
    UNUSED_ARG(S); /* unused input argument */
}


/***********************************************************
*   Linear interpolation of evenly spaced values
*   - yvalues are the known y-values (M elements, with a spacing dx from x=0)
*   - xvalues are the desired x-values (P elements)
*/
static void interp(real_T* yvalues, real_T* xvalues, real_T* yinterp, int M, int P, real_T dx){
	int i, lowindex;
	real_T ylow, yhi;
	for(i=0; i<P; i++){
		lowindex = (int)floor(xvalues[i]/dx);        
		ylow = yvalues[lowindex];
		yhi = yvalues[lowindex+1];
		yinterp[i] = (xvalues[i]-lowindex*dx)*(yhi-ylow)/dx+ylow;
	}	
}


/***********************************************************
*   Linear interpolation of evenly spaced values when one value is appended
*   to the start of the source vector (see interp())
*/
static void alpha_interp(real_T* yvalues, real_T* xvalues, real_T* yinterp, int M, int P, real_T dx, real_T alphaStart){
    int i, lowindex;
	real_T ylow, yhi;
    for(i=0; i<P; i++){
        if(xvalues[i] <= 0 && xvalues[i] < dx){
            yinterp[i] = xvalues[i]*(yvalues[0]-alphaStart)/dx+alphaStart;
        } else {
            lowindex = (int)floor((xvalues[i]-dx)/dx);
            ylow = yvalues[lowindex];
            yhi = yvalues[lowindex+1];
            yinterp[i] = (xvalues[i]-(lowindex+1)*dx)*(yhi-ylow)/dx+ylow;
        }
    }
}

/***********************************************************
*   Linear interpolation of evenly spaced values when one value is appended
*   to the end of the source vector (see interp())
*/
static void beta_interp(real_T* yvalues, real_T* xvalues, real_T* yinterp, int M, int P, real_T dx, real_T betaStart){
    int i, lowindex;
	real_T xlow, xhi, ylow, yhi;
    for(i=0; i<P; i++){
        if(xvalues[i] >= (M-1)*dx && xvalues[i] <= M*dx){
            yinterp[i] = (xvalues[i]-(M-1)*dx)*(betaStart-yvalues[M-1])/dx+yvalues[M-1];
        } else {
            lowindex = (int)floor(xvalues[i]/dx);        
            ylow = yvalues[lowindex];
            yhi = yvalues[lowindex+1];
            yinterp[i] = (xvalues[i]-lowindex*dx)*(yhi-ylow)/dx+ylow;
        }
    }
}

#ifdef MATLAB_MEX_FILE /* Is this file being compiled as a MEX-file? */
#include "simulink.c" /* MEX-file interface mechanism */
#else
#include "cg_sfun.h" /* Code generation registration function */
#endif
